(* This file belongs to the Secure Sessions Project 	      *)
(* http://www.msr-inria.inria.fr/projects/sec/sessions/       *)
(* MSR-INRIA Joint Centre, 2007    			      *)

(********************************************************************)
(* Session - Implementation                                         *)
(*                                                                  *)
(* graph.ml                                                         *)
(********************************************************************)
(* $Id: graph.ml 2847 2007-06-25 14:40:15Z denielou $ *)

(*#use "c:/Documents and Settings/t-ricarc/inria-msr-sec/lang-sec/session-types/src/syntax.ml";;*)
(*#use "/Users/ricardocorin/inria-msr-sec/lang-sec/session-types/src/syntax.ml";;*)

open Syntax 

let debug = Common.debug Common.debug_graph  "graph"

(* Regression tests material *)
let reg_1 s f x y = Common.regression1 debug
let reg_A = [("client","string",Send(0,[("Request","string"),
   Receive(1,[(("Ack","unit"),Goto 0); (("NAck","unit"),End 2)])]));
   ("server","string",Receive(0,[("Request","string"),
   Send(1,[(("Ack","unit"),Goto 0); (("NAck","unit"),End 3)])])) ]

(* Explanations ...

   We start from a session, ie 
   
   let pingpong : session =
   [("client","string",Send(0,[("Request","string"),
   Receive(1,[(("Ack","unit"),Goto 0); (("NAck","unit"),End 2)])]));
   ("server","string",Receive(0,[("Request","string"),
   Send(1,[(("Ack","unit"),Goto 0); (("NAck","unit"),End 3)])])) ] ;;

   and we want to construct several structures from it to both check
   properties and generate ML code.

   1) we need to generate the protocol graph as a list of edges to
   check the correctness properties and to generate the wired messages
   types and code.
   2) we need a flattened structure for each role to generate the
   corresponding proxy.

*)

(* type graph is a list of edges *)
type graph = (node * message * node) list
(* type assoc relates the nodes to the roles*)
type assoc = (node * string) list

(* Errors *)
exception Graph_error of string

(* Associate nodes to roles *)
let rec assoc_from_session : session -> assoc = 
  let rec assoc_from_role_type (r:string) = function
    | End n -> [n,r]
    | Goto _ -> []
    | Send (n,l) -> (n,r)::(assoc_from_msg_list r l)
    | Receive (n,l) -> (n,r)::(assoc_from_msg_list r l)
  and assoc_from_msg_list (r:string) = function 
      [] -> []
    | (_,rt)::q -> (assoc_from_role_type r rt)@(assoc_from_msg_list r q)
  in function
      [] -> []
    | (r,_,rt)::q -> (assoc_from_role_type r rt)@(assoc_from_session q);;




(* let _ = reg_1 "assoc_from_session" assoc_from_session reg_A 
   ([(0,"client");(1,"client");(2,"client");(0,"server"); (1,"server"); (3,
   "server")]) *)

(* Normalize the node substitution *)
let rec normalize na = 
  let rec isin (a,b) = function []->false | (c,d)::l -> (a=c && d=b) || (isin (a,b) l) in
   match na with
     | [] -> []
     | (a,b)::l ->  
	 if isin (a,b) l then normalize l else
	   begin
	     if List.mem_assoc a l 
	     then 
	       let newDest =   List.assoc a l in
		 debug (Printf.sprintf "normalizing a=%d newDest=%d b=%d" a newDest b);
		 (normalize ((newDest,b)::l)) 
	     else ((a,b)::normalize l) 
	   end

let alias (na:(node*node) list) = function n ->
  if List.mem_assoc n na 
  then List.assoc n na
  else n

(* Transforms a session graph into a list of edges *)
let flatten : role_type -> role_type list =
    let rec flatten_aux : role_type list -> role_type list = 
      let rec aux = function
	  [] -> raise (Graph_error "flatten")
	| [msg,s] -> begin 
	      match s with
		  End _ -> [msg,s]
		| _ -> [msg,Goto (node_of s)]
	  end
	|(msg,s)::q -> (match s with
	      End _ -> (msg,s)
	    | _ -> (msg,Goto (node_of s)))::(aux q) 
      and aux_top = function
	  [] -> []
	|(_,s)::q -> (match s with
	      End _ -> []
	    | _ -> [s])@(aux_top q)
      in function
	  [] -> []
	| s::q -> begin 
	      match s with
		  End _ -> []
		| Goto _ -> []
		| Send (n,mrtl) -> (Send (n,aux mrtl))::(flatten_aux (aux_top mrtl))
		| Receive (n,mrtl) -> (Receive (n,aux mrtl))::(flatten_aux (aux_top mrtl))
	  end @ (flatten_aux q)
    in
      function s -> flatten_aux [s] 

(* Selects only a node from a flatten version of a protocol *)
let rec old_filter_node (n:node) : role_type list -> role_type = function
    [] -> raise (Graph_error "old_filter_node")
  | (Send (i,l))::q -> if i = n then Send (i,l) else old_filter_node n q
  | (Receive (i,l))::q -> if i = n then Receive (i,l) else old_filter_node n q
  | _ -> raise (Graph_error "old_filter_node")

(* Selects only a node from a flatten version of a protocol *)
let rec filter_node acc (n:node) = function
    [] -> raise (Graph_error "filter_node")
  | (Send (i,l))::q -> if i = n then (i,l),(acc@q) else filter_node (acc@[(Send (i,l))]) n q
  | _ -> raise (Graph_error "filter_node")

(* Selects only a node from a flatten version of a protocol *)
let rec find_node (n:node) : role_type list -> bool = function
    [] -> false
  | (Send (i,l))::q -> (i = n) || (find_node n q)
  | _ -> raise (Graph_error "find_node")

(* Test for End of protocol *)
let is_End (rt:role_type) : bool =
  match rt with
    | End _ -> true
    | _ -> false

(* Presence test *)
let rec isin (msg:message) : (message * role_type) list -> bool = function 
    [] -> false
  | (m,_)::r -> (msg=m) || isin msg r

(* Strict Absence test *)
let rec notin (msg:message) : (message * role_type) list -> bool = function 
    [] -> true
  | (m,_)::r -> if (msg=m) then false else notin msg r

(* Strict Absence test *)
let rec notin_hard (msg:message) : (message * role_type) list -> bool = function 
    [] -> true
  | (m,_)::r -> 
      if (msg=m) 
      then raise (Graph_error ("notin error with message "^(fst m))) 
      else notin_hard msg r

(* Unique Presence test *)
let rec isin_once (msg:message) : (message * role_type) list -> bool = function 
    [] -> false
  | (m,_)::r -> 
(*      let _ = debug (Printf.sprintf "is %s present only once?" (fst m)) in *)
      if (msg=m) then notin_hard msg r else isin_once msg r

(* Inclusion test *)
let rec included 
    (msg_list_1: (message * role_type) list)
    (msg_list_2: (message * role_type) list) = 
  match msg_list_1 with 
    | [] -> true
    | (m,_)::b -> (isin m msg_list_2) && (included b msg_list_2)

(* Void intersection test *)
let rec void_intersect
    (msg_list_1: (message * role_type) list)
    (msg_list_2: (message * role_type) list) = 
  match msg_list_1 with 
    | [] -> true
    | (m,_)::b -> 
(*        let _ = debug (Printf.sprintf "%s shouldn't be present anymore" (fst m)) in *)         
        (notin_hard m msg_list_2) && (void_intersect b msg_list_2)

(* Inclusion test with strictness constraint *)
let rec included_once 
    (n:node)
    (msg_list_1: (message * role_type) list)
    (msg_list_2: (message * role_type) list) = 
  try begin
  match msg_list_1 with 
    | [] -> true
    | (m,_)::b -> 
        let _ = debug (Printf.sprintf "is %s present in the expected msg of node %i?" (fst m) n) in
        if (isin_once m msg_list_2) 
        then 
          let _ = debug (Printf.sprintf "%s is present once" (fst m)) in
          (included_once n b msg_list_2)
        else 
          let _ = debug  (Printf.sprintf "%s is not present" (fst m)) in
          let _ = (void_intersect b msg_list_2) in
          false
    end
  with _ -> raise (Graph_error "included_once")
(*
  let role_list = List.map snd pingpong in
  let action_list = List.concat (List.map flatten role_list) in
  let rec split sl rl = function
      [] -> (sl,rl)
    | a::q -> 
        begin
          match a with
            | Send _ -> split (sl@[a]) rl q
            | Receive _ -> split sl (rl@[a]) q
            | _ -> raise (Graph_error "split")
        end in
  split [] [] action_list 
*)


    
(**********************)
(* Graph Manipulation *)
(**********************)

(* From a graph we select the edges starting from node k *)
let rec select k : graph -> graph * graph = function
    [] -> [],[]
  | (i,(c,t),j)::q -> 
      let a,b = select k q in
      if k = i then ((i,(c,t),j)::a,b) else (a,(i,(c,t),j)::b)
        
(* From a graph we get the unique edge determined by a node and a message *)
let rec extract k m : graph -> node*message*node = function
    [] -> assert false
  | (i,p,j)::q -> if i=k && p=m then (i,p,j) else extract k m q

(* Sorts by common starting nodes *)
let rec sort_by_send : graph -> (node * graph) list = function
    [] -> []
  | (i,(c,t),j)::q -> 
      let a,b = select i ((i,(c,t),j)::q) in
      (i,a)::(sort_by_send b)

(* Goes in depth through the graph and applies a function on the way *)
let depth_check (g:graph) f (initial:node) =
  let rec check acc k =
    let sl,_ = select k g in
    List.fold_left 
      (function b -> function (_,(_,_),j) -> 
        b && (List.mem j acc || (f j && check (j::acc) j))) true sl
  in
  check [] initial

  
(**************************)
(* Verification functions *)
(**************************)

(* Property 1:
   By construction the graphs' edges have different source and target roles, but
   we can check it anyway *)
let rec check_no_self_send (role_assoc:assoc) (g:graph) =
  match g with
      [] -> ()
    | (i,(s,t),j)::q -> 
        if (List.assoc i role_assoc = List.assoc j role_assoc)
        then begin Printf.printf 
            "The message %s carrying a value of type %s is internal to %s.\n" 
            s t (List.assoc i role_assoc) ;
          raise (Graph_error "The graph does not satisfy the property 1.")
        end
        else (check_no_self_send role_assoc q)

(* Property 2:
   We check the unicity of labels *)
let check_unique_target (role_assoc:assoc) (g:graph) =
  let sorted_graph = 
    List.sort (function (_,(s,_),_) -> function (_,(t,_),_) -> compare s t) g in 
  let rec check_labels (i,c,j) = function
      [] -> ()
    | (k,(s,_),l)::q -> 
        if c = s 
        then 
          if (j<>l) || (List.assoc i role_assoc <> List.assoc k role_assoc)
          then begin Printf.printf 
              "Two different messages have label %s: from %d to %d and from %d to %d.\n" 
              s i j k l ;
            raise (Graph_error "The graph does not satisfy the property 2.")
          end
          else (check_labels (i,c,j) q)
        else check_labels (k,s,l) q
  in
  match sorted_graph with
      [] -> assert false
    | (k,(s,_),l)::q -> check_labels (k,s,l) q
          
let rec sprint_list = function
    [] -> ""
  | [s] -> s
  | s::q -> s^","^(sprint_list q)

(* Property 3:
   We check the No Blind Fork property. *)
(* Since we enforce that each send only send to one role, this property is
   already enforced for less than 3 roles *)
let check_no_blind_fork (role_assoc:assoc) (g:graph) =
  let rec check_width origin (visi,visj) roles i j (sli,slj) = 
    match (sli,slj) with
        [],[] -> true
      | [],(j,n,k)::r ->  let _ = debug 
          (Printf.sprintf "Prop 3: from %d,%d, looking towards %d" i j k) 
        in ((i=j) || (List.mem (j,n,k) visj) (*|| (List.mem (List.assoc k role_assoc) roles) *)
             || (check_depth origin (visi,(j,n,k)::visj) ((List.assoc k role_assoc)::roles) i k))
        && (let _ = debug (Printf.sprintf "Prop 3: from %d,%d, looking after %d" i j k) in 
            (List.assoc j role_assoc) = origin || check_width origin (visi,visj) roles i j ([],r))
      | (i,m,k)::q,[] -> let _ = debug 
          (Printf.sprintf "Prop 3: from %d,%d, looking towards %d" i j k)
        in ((i=j) || (List.mem (i,m,k) visi) (*|| (List.mem (List.assoc k role_assoc) roles) *)
             || (check_depth origin ((i,m,k)::visi,visj) ((List.assoc k role_assoc)::roles) k j))
        && (let _ = debug (Printf.sprintf "Prop 3: from %d,%d, looking after %d" i j k) in 
            (List.assoc i role_assoc) = origin || check_width origin (visi,visj) roles i j (q,[]))
      | (i,m,k)::q,(j,n,l)::r -> let _ = debug 
          (Printf.sprintf "Prop 3: Width =>from %d,%d, looking at %d, %d" i j k l)
        in (
          if ((List.mem (List.assoc k role_assoc) roles) 
               || (List.mem (List.assoc l role_assoc) roles) 
               || (List.assoc k role_assoc) = (List.assoc l role_assoc))
          then
            ((List.assoc i role_assoc) = origin || (List.assoc j role_assoc) = origin
                || (List.mem (i,m,k) visi) || (List.mem (j,n,k) visj)
                || check_depth origin ((i,m,k)::visi,(j,n,l)::visj) roles k l)
          else begin
            Printf.printf "There is a forking attack against %s at node %d or %s at node %d;\n" 
              (List.assoc k role_assoc) k (List.assoc l role_assoc) l ;flush_all ();
            false
          end)
        && (let _ = debug (Printf.sprintf 
                              "Prop 3: from %d,%d, finished %d,%d, absorbing %d" i j k l j) in
            (List.assoc j role_assoc) = origin || check_width origin (visi,visj) (roles) i j (sli,r))
        && (let _ = debug (Printf.sprintf "Prop 3: from %d,%d, finished %d,%d, absorbing %d" i j k l i) in
            (List.assoc i role_assoc) = origin || check_width origin (visi,visj) (roles) i j (q,slj))
  and check_depth origin visited roles i j =
    let _ = debug (Printf.sprintf "Prop 3: Depth => looking at nodes %d and %d knowing %s"
                      i j (sprint_list roles)) in
    let sli,_ = select i g in
    let slj,_ = select j g in
    if ((List.mem (List.assoc i role_assoc) roles) ||
      (List.mem (List.assoc j role_assoc) roles) ||
      (List.assoc i role_assoc = List.assoc j role_assoc))
    then
      (check_width origin visited ((List.assoc i role_assoc)::roles) i j (sli,slj))
    else begin
      Printf.printf "There is a forking attack against %s or %s at nodes %d and %d;\n" 
        (List.assoc i role_assoc) (List.assoc j role_assoc) i j ;flush_all ();
      false
    end
  in
  List.for_all
    (function (i,(m,_),j) -> 
      let _ = debug (Printf.sprintf "Prop 3: Initial from node %d to node %d !!\n" i j) in
      if check_depth (List.assoc i role_assoc) ([],[]) [] j j
      then true
      else begin
        Printf.printf 
          "this attack starts at node %d by a message %s from %s to the node %d.\n" 
          i m (List.assoc i role_assoc) j;
        raise (Graph_error "The graph does not satisfy the property 3.")
      end)
    g


let check_properties (role_assoc:assoc) (g:graph) =
  let _ = debug (Printf.sprintf "Checking the properties of the session.") in
  let _ = (check_no_self_send role_assoc g) in
  let _ = (check_unique_target role_assoc g) in
  let _ = (check_no_blind_fork role_assoc g) in
  ()


(********************************************************)
(* Construction of the graph from the syntactic session *)
(********************************************************)

let graph_from_session (s:session) =
  
  (* Global definitions *)
  let role_assoc = assoc_from_session s in
  let role_list = List.map (fun (_,_,r)->r) s in
  let action_list = List.concat (List.map flatten role_list) in
  
  (* Splits a list of actions into Sends and Receives *)
  let rec split sl rl = function
      [] -> (sl,rl)
    | a::q -> begin match a with
        | Send _ -> split (sl@[a]) rl q
        | Receive _ -> split sl (rl@[a]) q
        | _ -> raise (Graph_error "split")
      end in
  let send_list,receive_list = split [] [] action_list in
  
  (* For each sent message we look for a reception *)
  let rec remove_role assoc p = function
      [] -> []
    | (Send (n,lm))::q -> 
        (if List.assoc n assoc = p then [] else [Send (n,lm)])@(remove_role assoc p q)
    | _ -> raise (Graph_error "remove role failure") in
  let rec find_only_one msg_list = function
      [] -> raise (Graph_error "find_only_one but found none")
    | (Receive (n,lm))::b -> 
        if included_once n msg_list lm 
        then (find_none msg_list b ; (n,lm))
        else find_only_one msg_list b
    | _ -> raise (Graph_error "find_only_one but found more")
  and find_none msg_list = function 
      [] -> ()
    | (Receive (n,lm))::b -> 
        if void_intersect msg_list lm 
        then find_none msg_list b 
        else raise (Graph_error "find_none: receive compromised")
    | _ -> raise (Graph_error "find_none: wrong split") in
  let node_aliases = ref [] in

  (* From each sending node we look for Receives that can get the messages *) 
  let rec build_from_send acc send_list recv_list (i:node) = 
    let _ = debug (Printf.sprintf "Looking for a receive corresponding to node %d ... " i) in
    (* We find the messages sent from node i *)
    let ((i,msg_sent_from_i),_) = filter_node [] i send_list in
    (* We look in rl for a receive that has all its messages included in
     * msg_sent_from_i *)
    (* This enforces the condition that from a given node all messages have to
     * be to the same role *)
    let j,msg_recv_in_j = 
      try 
        find_only_one msg_sent_from_i recv_list 
      with
          Graph_error s -> raise (Graph_error (s^" from node "^(string_of_int i)^" "))
    in
    let _ = node_aliases := (j,i)::!node_aliases in (* We record the correspondence *)
    let _ = debug (Printf.sprintf "Adding alias %i <- %i" i j) in
    let _ = debug (Printf.sprintf "-> Receive found for messages sent at node %d ! " i) in
    (* Now we can add the corresponding edges to the graph *)
    build_edges acc send_list recv_list i msg_sent_from_i j msg_recv_in_j

  and build_edges graph send_list rl i msg_sent_from_i j msg_recv_in_j =
    (* Builds the edges of the graph starting in i and j *)
       match msg_sent_from_i with
         | [] -> graph (* We looked at all send transitions *)
      | (m,send_goto)::remaining_msg -> 
          begin
            let receive_goto = List.assoc m msg_recv_in_j in
            let next_node = node_of receive_goto in
            let continue,new_graph = (* do we look at the continuation of the
                                        current edge, and updated graph *)
              if List.mem (i,m,next_node) graph 
              then false,graph (* edge already added *)
              else (true,(i,m,next_node)::graph) (* addition of the edge to the graph *)
            in match (continue,is_End send_goto,is_End receive_goto,find_node next_node send_list) with
              | true,false,false,true -> 
                  (* we continue to look through the current edge, plus we proceed with the
                     remaining parallel edges *)
                  (build_from_send (build_edges (new_graph) send_list rl i remaining_msg j msg_recv_in_j) send_list rl next_node)
              | true,true,false,true ->  
                  (* The role who is sending the message is terminating: we remove all nodes
                     corresponding to this role *)
                  let new_sl = remove_role role_assoc (List.assoc i role_assoc) send_list in
                  (build_from_send (build_edges (new_graph) send_list rl i remaining_msg j msg_recv_in_j) new_sl rl next_node)
              | false,_,false,true -> (* This edge is already in the graph *)
                  (build_edges new_graph send_list rl i remaining_msg j msg_recv_in_j)
              | _,true,true,false -> (* This edge is an end for both principals *)
                  (build_edges new_graph send_list rl i remaining_msg j msg_recv_in_j)
              | _ -> raise (Graph_error "incoherent Endings") 
                  (* Any other possibility is wrong *)
          end 
  in
  let _ = debug (Printf.sprintf "Graph generation started.") in
  let g = build_from_send [] send_list receive_list (node_of (List.hd
    send_list)) in
  let _ = debug (sprint_session s) in
  let _ = debug (Printf.sprintf "Graph generation completed.") in
  let _ = check_properties role_assoc g in
  (g,!node_aliases)


(***********************************)
(* Compute the visibility function *)
(***********************************)

let compute_visibility (g:graph) (role_assoc: (node*string)list) 
    : node -> message -> (string * message) list list =
  let _ = debug (Printf.sprintf "Visibility calculation started.") in
  let rec isin (a,b)  = function []->false | (c,d)::l -> (a=c && b=d) || isin (a,b) l in
  let revg = List.map (fun (i,m,j) -> (j,m,i)) g in
  let sorted = sort_by_send revg in
  let rec aux role current role_list seen_nodes = 
    if (isin (role, current) seen_nodes) then [[]] else
    if not (List.mem_assoc current sorted)
    then [[]]
    else
      let le = List.assoc current sorted in
	List.concat 
	  (List.map 
	      (fun (_,m,n) -> 
		let r = List.assoc n role_assoc in
(*		  Printf.printf "Entered aux. n: %d current: %d r: %s\n" n  current r;*)
		  if r = role then [[]] else
		    if List.mem r role_list then aux role n role_list ((role,current)::seen_nodes)
		    else
		      List.map (fun x -> (r,m)::x) (aux role n (r::role_list) ((role,current)::seen_nodes))
	      )
      le)
  in
  let _ = debug (Printf.sprintf "Visibility function completed.") in
  fun (n:node) (m:message) ->
    (*	debug (Printf.sprintf "Called in n: %d" n);*)
    let (n,m,p) = extract n m revg in
    (*	debug (Printf.sprintf "After extract: n: %d p: %d" n p);*)
    let ro = List.assoc n role_assoc in
    let r = List.assoc p role_assoc in
    debug (Printf.sprintf "ro: %s r: %s " ro r);
    List.map (fun x -> (r,m)::x) (aux ro p [r] [])



(*********************************************)
(* Garbage for history and forgotten stuff : *)
(*********************************************)
(* let check_no_blind_fork_old (role_assoc:assoc) (g:graph) =
  (*  let rec check_first i roles =
    let sli,_ = select i g in
    let rec aux sli = 
      match sli with
          [] -> true
        | [(_,_,_)] -> true
        | (_,_,k)::(_,_,l)::q ->
          (List.mem (List.assoc k role_assoc) roles) ||
      ((check_depth ((List.assoc k role_assoc)::roles) k j) && 
                (check_width i j roles (q,[])))
      |(_,_,k)::q,(_,_,l)::r ->
       ((List.assoc k role_assoc) = (List.assoc l role_assoc)) &&
         (check_width i j roles (sli,r)) &&
      (check_width i j roles (q,slj)) && true *)
  let rec check_width i j roles (sli,slj) = 
    match (sli,slj) with
        [],[] -> true
      | [],(_,_,k)::q -> 
          let _ = debug (Printf.sprintf "Prop 3: from %d,%d, looking towards %d"
                            i j k) in
          (List.mem (List.assoc k role_assoc) roles || i = j) ||
            ((check_depth ((List.assoc k role_assoc)::roles) i k) && 
                (check_width i j roles ([],q)))
      | (_,_,k)::q,[] -> 
          let _ = debug (Printf.sprintf "Prop 3: from %d,%d, looking towards %d"
                            i j k) in
          (List.mem (List.assoc k role_assoc) roles || i = j) ||
            ((check_depth ((List.assoc k role_assoc)::roles) k j) && 
                (check_width i j roles (q,[])))
      |(_,_,k)::q,(_,_,l)::r ->
         let _ = debug (Printf.sprintf "Prop 3: from %d,%d, looking at nodes %d and %d" i j k l) in
         ((List.mem (List.assoc k role_assoc) roles) ||
             (List.mem (List.assoc l role_assoc) roles) ||
             ((((List.assoc k role_assoc) = (List.assoc l role_assoc)) && check_depth ((*(List.assoc k role_assoc)::*)roles) k l) &&
         (let _ = debug (Printf.sprintf "Prop 3: from %d,%d, looking towards %d"
                            i j l) in
          check_width i j roles (sli,r)) &&
         (let _ = debug (Printf.sprintf "Prop 3: from %d,%d, looking towards %d"
                            i j k) in
          check_width i j roles (q,slj))))
  and check_depth roles i j =
    let _ = debug (Printf.sprintf "Prop 3: looking at nodes %d and %d" i j) in
    let sli,_ = select i g in
    let slj,_ = select j g in
    (List.mem (List.assoc i role_assoc) roles) ||
      (List.mem (List.assoc j role_assoc) roles) ||
      ((List.assoc i role_assoc = List.assoc j role_assoc) &&
          (check_width i j ((List.assoc i role_assoc)::roles) (sli,slj)))
  in
  List.for_all 
    (function (i,_,j) -> 
      let _ = debug (Printf.sprintf "Prop 3: Initial from node %d to node %d" i j) in
      check_depth [(*List.assoc j role_assoc;List.assoc i role_assoc*)] j j) g
  *)
